import datetime
import json

from django.contrib import messages
from django.contrib.admin.utils import NestedObjects
from django.core.exceptions import ImproperlyConfigured
from django.db import router
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseBadRequest, HttpResponseNotAllowed
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.html import escape, escapejs
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView, CreateView, UpdateView, DeleteView

from . import forms
from . import gas_settings
from .sites import site


class AjaxCommandsMixin:
    def post(self, request, *args, **kwargs):
        if 'command' in self.request.POST:
            command_processor = getattr(self, 'do_{0}'.format(self.request.POST['command']), None)
            if command_processor is not None:
                return command_processor()
            else:
                return HttpResponseBadRequest()
        else:
            handler = getattr(super(), 'post', self.http_method_not_allowed)
            return handler(request, *args, **kwargs)

    def render_json(self, data):
        return HttpResponse(json.dumps(data), content_type='application/json')


class GASMixin:
    base_role = 'admins'
    roles = set()
    base_template = 'gas/base.html'
    continue_url = None
    header_title = ''
    title = ''
    help_text = ''
    success_message = _("Operation successful.")
    breadcrumbs = []
    actions = None

    def dispatch(self, *args, **kwargs):
        user = self.request.user
        roles = set(self.roles)
        roles.add(self.base_role)
        access_denied = (
            not user.is_authenticated or (
                 not user.is_superuser 
                 and user.user_roles.filter(role__in=roles).count() == 0
            )
        )
        if access_denied:
            return HttpResponseRedirect(reverse('gas:login') + '?next={}'.format(self.request.path))
        return super().dispatch(*args, **kwargs)

    def form_valid(self, form):
        response = super().form_valid(form)
        if 'save_and_continue' in self.request.POST:
            response = HttpResponseRedirect(self.get_continue_url())
        messages.add_message(self.request, messages.SUCCESS, self.get_success_message())
        return response

    def get_success_message(self):
        return self.success_message

    def get_continue_url(self):
        if self.continue_url:
            # Forcing possible reverse_lazy evaluation
            url = force_text(self.continue_url)
            return url
        else:
            raise ImproperlyConfigured("No URL to redirect to. Provide a continue_url.")

    def get_header_title(self):
        " Contents for the <title> tag. "
        return self.header_title or self.title

    def get_title(self):
        " Contents for page title. "
        return self.title

    def get_help_text(self):
        " Contents for page help. "
        return self.help_text

    def get_breadcrumbs(self):
        " Returns a list of (url, label) tuples for the breadcrumbs "
        return self.breadcrumbs

    def get_actions(self):
        return self.actions or []

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        css = gas_settings.MEDIA['css']
        js = gas_settings.MEDIA['js']
        if gas_settings.EXTRA_MEDIA:
            css = css + gas_settings.EXTRA_MEDIA.get('css', [])
            js = js + gas_settings.EXTRA_MEDIA.get('js', [])
        ctx.update({
            'base_template': self.base_template,
            'header_title': self.get_header_title(),
            'title': self.get_title(),
            'help_text': self.get_help_text(),
            'breadcrumbs': self.get_breadcrumbs(),
            'actions': self.get_actions(),
            'gas_title': gas_settings.TITLE,
            'logo_static_url': gas_settings.LOGO,
            'css': css,
            'js': js,
        })
        return ctx


class GASListView(GASMixin, ListView):
    """  ListView, permite indicar un formulario para filtrar contenido. """
    filter_form_class = None

    def get_filter_form(self):
        if self.filter_form_class is None:
            return None
        data = self.request.GET
        return self.filter_form_class(data=data)

    def filter_queryset(self, qs):
        self.filter_form = self.get_filter_form()
        if self.filter_form is not None and self.filter_form.is_valid():
            return self.filter_form.filter(qs)
        else:
            return qs

    def get_queryset(self):
        qs = super().get_queryset()
        return self.filter_queryset(qs)

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx.update({
            'filter_form': self.filter_form,
        })
        return ctx


class GASCreateView(GASMixin, CreateView):
    template_name = 'gas/base_form.html'

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        is_popup = "_popup" in self.request.GET
        ctx.update({
            'is_popup': is_popup,
        })
        return ctx

    def form_valid(self, form):
        response = super().form_valid(form)
        if '_popup' in self.request.POST:
            return HttpResponse(
                '<!DOCTYPE html><html><head><title></title></head><body>'
                '<script type="text/javascript">opener.dismissAddAnotherPopup(window, "%s", "%s");</script></body></html>' %
                # escape() calls force_text.
                (escape(self.object.pk), escapejs(self.object)))
        else:
            return response


class GASUpdateView(GASMixin, UpdateView):
    """ Same as Django's UpdateView, defined only for completeness. """
    template_name = 'gas/base_form.html'


class GASDeleteView(GASMixin, DeleteView):
    template_name = "gas/delete_confirmation.html"
    confirmation_text = _("Are you sure you want to delete {object}?")
    deleted_text = _("{object} deleted.")

    def get_confirmation_text(self):
        return self.confirmation_text.format(object=self.object)

    def get_deleted_text(self):
        return self.deleted_text.format(object=self.object)

    def get_deleted_objects(self):
        using = router.db_for_write(self.model)
        collector = NestedObjects(using=using)

        def format_callback(obj):
            opts = obj._meta
            return '%s: %s' % (capfirst(opts.verbose_name), obj)

        collector.collect([self.object])
        model_count = {model._meta.verbose_name_plural: len(objs) for model, objs in collector.model_objs.items()}
        return collector.nested(format_callback), model_count

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)

        deleted_objects, deleted_model_count = self.get_deleted_objects()

        ctx.update({
            'confirmation_text': self.get_confirmation_text(),
            'cancel_url': self.get_success_url(),
            'deleted_objects': deleted_objects,
            'deleted_model_count': deleted_model_count,
        })
        return ctx

    def delete(self, request, *args, **kwargs):
        response = super().delete(request, *args, **kwargs)
        messages.add_message(request, messages.SUCCESS, self.get_deleted_text())
        return response



import json
from urllib.parse import urlencode

from django import template
from django.utils.html import mark_safe

from ..sites import site

register = template.Library()


@register.inclusion_tag('gas/tags/navigation.html')
def show_navigation(user):
    user_roles = set(user.user_roles.values_list('role', flat=True))
    if user.is_superuser:
        user_roles.add('admins')
    menu = site.get_root_menu(user_roles)
    menu.sort()
    return {
        'menu': menu,
        'user_roles': user_roles,
    }


@register.filter
def get_children(entry, roles):
    return sorted(
        [
            child for child in entry.children.values()
            if child.roles.intersection(roles)
        ],
    )


@register.filter()
def display_boolean(value):
    if bool(value):
        render = '<i class="fas fa-check"></i>'
    else:
        render = '<i class="fas fa-times"></i>'
    return mark_safe(render)


@register.simple_tag
def url_replace(request, **kwargs):
    query = request.GET.dict()
    query.update(kwargs)
    return urlencode(query)


@register.inclusion_tag("gas/tags/pagination.html")
def pagination(request, page):
    return {
        'request': request,
        'page': page,
    }


@register.filter
def to_json(data):
    return mark_safe(json.dumps(data))



from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from django.urls import reverse

from gas.sites import GASSite, site


class SiteTestCase(TestCase):
    def test_register_role(self):
        test_site = GASSite()

        # new role registered without error
        test_site.register_role('test', 'test role')

        # another new role registered without error
        test_site.register_role('test2', 'another test role')

        # repeated role raises error
        with self.assertRaises(ImproperlyConfigured):
            test_site.register_role('test', 'repeated test role')

    def test_register_urls(self):
        test_site = GASSite()

        # new prefix registered without error
        test_site.register_urls('test_prefix', 'gas.gas.urls')

        # another new prefix registered without error
        test_site.register_urls('test_prefix2', 'gas.gas.urls')

        # repeated prefix raises error
        with self.assertRaises(ImproperlyConfigured):
            test_site.register_urls('test_prefix', 'gas.gas.urls')

    def test_register_menu(self):
        test_site = GASSite()
        test_site.register_role('test_admins', 'test administrators')

        # Menu entry with non existent parent raises error
        with self.assertRaises(ImproperlyConfigured):
            test_site.register_menu(
                parent='nonexistent',
                name='test_roles',
                label="Roles",
                icon="",
                url="gas:role_list",
                roles=('test_admins',),
            )

        # new entry without parent without error
        test_site.register_menu(
            name='sample_app',
            label="Sample app",
            icon="",
            url=None,
            roles=('test_admins',),
        )

        # new entry with known parent without error
        test_site.register_menu(
            parent='sample_app',
            name='test_roles',
            label="Roles",
            icon="",
            url="gas:role_list",
            roles=('test_admins',),
        )
        # new entry with duplicated name raises error
        with self.assertRaises(ImproperlyConfigured):
            test_site.register_menu(
                name='sample_app',
                label="Sample app 2",
                icon="",
                url=None,
                roles=('test_admins',),
            )

        # new entry with no role
        test_site.register_menu(
            parent='sample_app',
            name='no_roles',
            label="No Roles",
            icon="",
            url="gas:role_list",
        )

    def test_menu_ordering(self):
        test_site = GASSite()
        admin_roles = set(('admins',))

        # With no explicit order, sort by label
        test_site.register_menu(name='c', label="c")
        test_site.register_menu(name='b', label="b")
        test_site.register_menu(name='a', label="a")
        entries = [
            entry.name
            for entry in sorted(test_site.get_root_menu(admin_roles))
        ]
        self.assertEqual(entries, ['a', 'b', 'c'])

        # Explicit order takes preference
        test_site.register_menu(name='z', label="z", order=1)
        entries = [
            entry.name
            for entry in sorted(test_site.get_root_menu(admin_roles))
        ]
        self.assertEqual(entries, ['z', 'a', 'b', 'c'])

        # Explicit order inside parent don't change the order
        test_site.register_menu(name='y', label="y", order=2, parent='a')
        entries = [
            entry.name
            for entry in sorted(test_site.get_root_menu(admin_roles))
        ]
        self.assertEqual(entries, ['z', 'a', 'b', 'c'])

        # With same explicit order, sort by label
        test_site.register_menu(name='x', label="x", order=1)
        entries = [
            entry.name
            for entry in sorted(test_site.get_root_menu(admin_roles))
        ]
        self.assertEqual(entries, ['x', 'z', 'a', 'b', 'c'])

    def test_menu_visibility(self):
        test_site = GASSite()
        admin_roles = set(('admins',))
        staff_roles = set(('staff',))

        test_site.register_menu(name='a', label='a', roles=admin_roles)
        test_site.register_menu(name='b', label='b', roles=staff_roles)

        # Admins gets all menu
        entries = [
            entry.name
            for entry in test_site.get_root_menu(admin_roles)
        ]
        self.assertEqual(entries, ['a', 'b'])

        # Staff gets filtered menu
        entries = [
            entry.name
            for entry in test_site.get_root_menu(staff_roles)
        ]
        self.assertEqual(entries, ['b'])


class SampleAppIntegrationTest(TestCase):
    def test_gas_autodiscover(self):
        # urls registered
        reverse('gas:user_list')

        # menu registered
        self.assertIn(
            'users',
            site._registry['menu'],
        )

        # role registered
        self.assertIn(
            'admins',
            site._registry['roles'],
        )



from collections import namedtuple

from django.urls import include, re_path
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import autodiscover_modules
from django.utils.translation import gettext_lazy as _


class Entry:
    def __init__(self, name, label, icon=None, url=None, roles=None, parent=None, order=None):
        self.name = name
        self.label = label
        self.icon = icon
        self.url = url
        self.roles = roles
        self.children = dict()
        self.parent = parent
        self.order = order if order is not None else float("inf")

    def __lt__(self, other):
        if isinstance(other, Entry):
            return (self.parent, self.order, self.label) < (other.parent, other.order, other.label)
        return super().__lt__(other)


class GASSite(object):
    base_role = 'admins'

    def __init__(self):
        self._registry = {
            'menu': {},
            'urls': {},
            'roles': {},
        }
        self.autodiscover_done = False

    def register_role(self, name, description):
        if name in self._registry['roles']:
            raise ImproperlyConfigured("Role already registered")
        self._registry['roles'][name] = description

    def register_urls(self, prefix, urls):
        """
            Register urls with a prefix. Compared to normal urls, these are not
            exposed when gas is not active.
        """
        if prefix in self._registry['urls']:
            raise ImproperlyConfigured("Prefix {0} already in use".format(prefix))
        self._registry['urls'][prefix] = urls

    def register_menu(self, name, label, url=None, icon=None, roles=None, parent=None, order=None):
        if name in self._registry['menu']:
            raise ImproperlyConfigured("Menu entry '{0}' already registered.".format(name))

        if roles is None:
            roles = set()
        else:
            roles = set(roles)

        roles.add(self.base_role)

        entry = Entry(name, label, icon, url, roles, parent=parent, order=order)

        if parent:
            try:
                parent_entry = self._registry['menu'][parent]
            except KeyError:
                raise ImproperlyConfigured("Parent {} not registered.".format(parent))
            parent_entry.children[name] = entry

        self._registry['menu'][name] = entry

    @property
    def urls(self):
        return self.get_urls(), 'gas'

    @property
    def role_choices(self):
        return [
            (role, f"{role}: {description}")
            for role, description in self._registry['roles'].items()
        ]

    def get_role_description(self, role):
        return self._registry['roles'].get(role, '')

    def get_urls(self):
        urlpatterns = []
        for prefix, urls in self._registry['urls'].items():
            if prefix:
                urlpatterns.append(
                    re_path(r'^{0}/'.format(prefix), include(urls)),
                )
            else:
                urlpatterns.append(
                    re_path(r'', include(urls)),
                )
        return urlpatterns

    def get_root_menu(self, user_roles):
        entries = []
        for entry in self._registry['menu'].values():
            if entry.parent is not None:
                continue
            visible = user_roles.intersection(entry.roles)
            if not visible:
                continue
            entries.append(entry)
        return entries

    def autodiscover(self):
        if not self.autodiscover_done:
            autodiscover_modules('gas.config', register_to=site)
            self.autodiscover_done = True


site = GASSite()



from django import forms
from django import template
from django.conf import settings
from django.template.loader import render_to_string
from django.urls import reverse

register = template.Library()


def base_form_field(field, css=None, container_class='', add_another_url=None, field_template=None):
    if field.is_hidden:
        return str(field)


    classes = field.field.widget.attrs.get('class', '').split(' ')
    classes.append('form-control')
    if css:
        classes.append(css)

    if isinstance(field.field.widget, forms.Select):
        classes.append('select2')
    elif isinstance(field.field, forms.DateField):
        field.field.widget.input_type = 'date'

    if isinstance(field.field, forms.SplitDateTimeField):
        field.field.widget.widgets[1].attrs['placeholder'] = '00:00:00'
    else:
        field.field.widget.attrs['placeholder'] = field.label

    if field_template is None:
        if isinstance(field.field.widget, forms.CheckboxInput):
            field_template = 'gas/tags/forms/checkbox.html'
        else:
            field_template = 'gas/tags/forms/field.html'

    field.field.widget.attrs['class'] = " ".join(classes)
    return render_to_string(field_template, {
        'field': field,
        'add_another_url': add_another_url,
        'container_class': container_class,
    })


@register.simple_tag
def add_widget_attrs(field, **attrs):
    field.field.widget.attrs.update(attrs)
    return ''


@register.inclusion_tag('gas/tags/forms/errors.html')
def form_errors(form, container_class=''):
    return {
        'form': form,
        'container_class': container_class,
    }


register.simple_tag(base_form_field, name='form_field')



from django import forms
from django.contrib.auth.models import User
from django.utils.translation import gettext_lazy as _

from gas.sites import site


class BaseFilterForm(forms.Form):
    """
        Form for content filtering.

        The `filter` method must be replaced.
    """
    def filter(self, qs):
        """
            This function takes a queryset and returns another queryset, with
            the content filtered with the data in the form.
        """
        return qs


class UserForm(forms.ModelForm):
    roles = forms.MultipleChoiceField(
        choices=site.role_choices,
        label=_('roles'),
        required=False,
    )
    class Meta:
        model = User
        fields = (
            'first_name', 'last_name', 'username', 'email', 'is_active',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'roles' not in self.initial and self.instance.pk is not None:
            self.initial['roles'] = self.instance.user_roles.values_list('role', flat=True)

    def save(self, commit=True):
        obj = super().save(commit=False)

        old_save_m2m = self.save_m2m

        def save_m2m():
            old_save_m2m()
            roles = self.cleaned_data['roles']
            for role in roles:
                obj.user_roles.update_or_create(role=role)
            obj.user_roles.exclude(role__in=roles).delete()

        self.save_m2m = save_m2m

        if commit:
            obj.save()
            self.save_m2m()

        return obj


def split_datetime_field(form, field_name):
    old_field = form.fields[field_name]
    new_field = forms.SplitDateTimeField(
        label=old_field.label,
        label_suffix=old_field.label_suffix,
        required=old_field.required,
        help_text=old_field.help_text,
        initial=old_field.initial,
        validators=old_field.validators,
        localize=old_field.localize,
        disabled=old_field.disabled,
        input_date_formats=("%Y-%m-%d",),
        input_time_formats=("%H:%M", "%H:%M:%S"),
        widget=forms.SplitDateTimeWidget(
            date_format='%Y-%m-%d',
            time_format='%H:%M:%S',
            time_attrs={'placeholder': '00:00'},
        ),
    )
    new_field.widget.widgets[0].input_type = 'date'
    form.fields[field_name] = new_field
